/* original notes from Jason Summer's pngrewrite code */

/* pngrewrite
   version 1.2.1 - 27 Sep 2003

  A quick and dirty utility to reduce unnecessarily large palettes
  and bit depths in PNG image files.

  ** To compile in a unix-like environment, type something like:
  **   cc -o pngrewrite pngrewrite.c -lpng -lz -lm

  Web site: <http://www.pobox.com/~jason1/pngrewrite/>

  This program and source code may be used without restriction.

  Primary author: Jason Summers <jason1@pobox.com>

  Contributing authors:

     Wayne Schlitt
       - Write grayscale images when possible
           - Ability to sort palette by color usage
           - Improved find_pal_entry performance

     Soren Anderson
           - Changes to allow use in a unix-style pipeline
           - Improved compatibility with MinGW and Cygwin

  Pngrewrite will:

  * Remove any unused palette entries, and write a palette that is only as
    large as needed.

  * Remove (collapse) any duplicate palette entries.

  * Convert non-palette image to palette images, provided they contain no
    more than 256 different colors.

  * Write images as grayscale when possible.

  * Move any colors with transparency to the beginning of the palette, and
    write a tRNS chunk that is a small as possible.

  * Reduce the bit-depth (bits per pixel) as much as possible.

  All of this is just basic stuff that any respectable image writing program
  should always do automatically, but for some reason, many of them do not.

  Under no circumstances does pngrewrite change the actual pixel colors, or
  background color, or transparency of the image. If it ever does, that's a
  bug.

  --WARNING--
  This version of pngrewrite removes most extra (ancillary) information from
  the PNG file, such as text comments. Although this does make the file size
  smaller, the removed information may sometimes be important. For that
  reason, you should only use pngrewrite on image that were created by you.

  The only ancillary chunks that are NOT removed are:
     gAMA  - Image gamma setting
     sRGB  - srgb color space indicator
     bKGD and tRNS - Background color and transparency are maintained. The
        actual chunk may be modified according to the new color structure.

  If the original image was interlaced, the new one will also be interlaced.

  Pngrewrite will not work at all on images that have more than 256 colors.
  Colors with the same RGB values but a different level of transparency
  count as different colors. The background color counts as an extra color
  if it does not occur in the image.

  It will also not work at all on images that have a color depth of 16 bits,
  since they cannot have a palette.

  This is a very inefficient program. It is (relatively) slow, and uses tons
  of memory. To be specific, it uses about 5 bytes per pixel, no matter what
  the bit depth of the image is.

  This program is (hopefully) reasonably portable, and should compile
  without too much effort on most ANSI C compilers. It requires the libpng
  and zlib libraries.  */

/* pngrecolor: Copyright (C) 2004-2005 by Brian Lindholm:
  [1] Renamed from "pngrewrite" to "pngrecolor" at avoid confusion with
      original program.
  [2] Ran through GNU indent and removed unneeded braces.
  [3] Modified palette comparison routine to ensure a *unique* sorting.
  [4] Ensured that a non-zero exit code was returned on any sort of error.
  [5] Cleaned up some type-casts for a more proper compile.
  [6] Added command-line options "-n" and "-q" for no-interlacing and
      quiet output.
  These modifications may be used without restriction.  */


#define PNGRECOLORVERSION "1.2.1"

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined( WIN32 ) && !defined (__GNUC__)
#include <search.h>
#include <io.h>
#include <fcntl.h>
#endif

#if defined( WIN32 ) && defined (__GNUC__)
#  include <fcntl.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#define OPTEND -1
#else
#define OPTEND EOF
#endif

#include "png.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef __MINGW32__
extern int getopt (int argc, char * const *argv, const char *optstring);
extern char *optarg;
extern int optind;
#endif

#ifdef HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

typedef struct
{
  unsigned char red;
  unsigned char green;
  unsigned char blue;
  unsigned char alpha;
  unsigned int count;
}
tmppal_t;

static tmppal_t tmppal[256];
static int pal_used, new_bit_depth, valid_gray;

static unsigned char *image1, *image2, **row_pointers;
static int rowbytes, channels;
static png_uint_32 width, height;
static int bit_depth, color_type, interlace_type, ori_pal_size;

static double image_gamma;
static tmppal_t bkgd;           /* RGB background color */
static unsigned char bkgd_pal;  /* new background color palette entry */
static int gray_trns, srgb_intent;

struct chunk_flags
{
  unsigned int has_gama:1;
  unsigned int has_bkgd:1;
  unsigned int has_srgb:1;
};
/*  initialize chunk-handling bitflags in read-png()  */
struct chunk_flags chnks;

#define PNG_READ_SIG_BYTES 7

static void
read_png (char *file_name, int verbose)
{
  png_structp png_ptr;
  png_infop info_ptr;
  png_colorp ori_pal;
  int ori_bpp;
  unsigned char header[PNG_READ_SIG_BYTES];
  int is_png = 0;

  FILE *fp;
  int i;
  png_color_16 *image_background;

  if ((fp = fopen (file_name, "rb")) == NULL)
    {
      fprintf (stderr, "pngrecolor error: can't open file %s\n", file_name);
      exit (1);
    }

  fread (header, 1, (int) PNG_READ_SIG_BYTES, fp);
  is_png = !png_sig_cmp (header, 0, PNG_READ_SIG_BYTES);
  if (!is_png)
    {
      fprintf (stderr, "pngrecolor error: not a PNG file, initial %d bytes "
        "are \"%s\"", PNG_READ_SIG_BYTES, header);
      exit (2);
    }

  chnks.has_gama = 0;
  chnks.has_bkgd = 0;
  chnks.has_srgb = 0;

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL)
    {
      fprintf (stderr, "pngrecolor error: can't create read_struct\n");
      (void) fclose (fp);
      exit (2);
    }

  png_set_sig_bytes (png_ptr, PNG_READ_SIG_BYTES);
  info_ptr = png_create_info_struct (png_ptr);
  if (info_ptr == NULL)
    {
      fprintf (stderr, "pngrecolor error: can't create info_struct\n");
      png_destroy_read_struct (&png_ptr, (png_infopp) NULL, (png_infopp) NULL);
      (void) fclose (fp);
      exit (2);
    }

  png_set_read_fn (png_ptr, fp, NULL);
  if (setjmp (png_jmpbuf (png_ptr)))
    {
      fprintf (stderr, "bizzare pngrecolor error: setjmp problem\n");
      png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp) NULL);
      (void) fclose (fp);
      exit (2);
    }

  png_init_io (png_ptr, fp);
  png_read_info (png_ptr, info_ptr);
  png_get_IHDR (png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
    &interlace_type, NULL, NULL);

  if (bit_depth > 8)
    {
      fprintf (stderr, "pngrecolor warning: 16 bits/sample - no conversion\n");
      png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp) NULL);
      (void) fclose (fp);
      exit (3);
    }

  if (color_type == PNG_COLOR_TYPE_PALETTE)
    {
      ori_pal_size = 0;
      ori_bpp = bit_depth;
      /* we only do this to get the palette size so we can print it */
      png_get_PLTE (png_ptr, info_ptr, &ori_pal, &ori_pal_size);
      if (verbose)
        fprintf (stdout, "original palette size of %s: %d, %d bpp\n",
          file_name, ori_pal_size, ori_bpp);
      png_set_expand (png_ptr);
    }
  else
    {
      /* figure out bits per pixel so we can print it */
      ori_bpp = bit_depth * (int) png_get_channels (png_ptr, info_ptr);
      if (verbose)
        fprintf (stdout, "original palette size of %s: [n/a], %d bpp\n",
          file_name, ori_bpp);
    }

  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
    png_set_expand (png_ptr);
  if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
    png_set_expand (png_ptr);
  /* if (bit_depth == 16)
     png_set_strip_16 (png_ptr); */
  if (color_type == PNG_COLOR_TYPE_GRAY ||
      color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
    png_set_gray_to_rgb (png_ptr);

  if (png_get_bKGD (png_ptr, info_ptr, &image_background))
    {
      chnks.has_bkgd = 1;
      bkgd.red = (unsigned char) image_background->red;
      bkgd.green = (unsigned char) image_background->green;
      bkgd.blue = (unsigned char) image_background->blue;
      bkgd.alpha = (unsigned char) 255;
    }

  if (png_get_gAMA (png_ptr, info_ptr, &image_gamma))
    chnks.has_gama = 1;

  if (png_get_sRGB (png_ptr, info_ptr, &srgb_intent))
    chnks.has_srgb = 1;

  png_read_update_info (png_ptr, info_ptr);
  rowbytes = (int) png_get_rowbytes (png_ptr, info_ptr);
  channels = (int) png_get_channels (png_ptr, info_ptr);

  if (channels < 3 || channels > 4)
    {
      fprintf (stderr, "internal pngrecolor error: channels = %d\n", channels);
      png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp) NULL);
      (void) fclose (fp);
      exit (4);
    }

  if (verbose)
    fprintf (stdout, "Image size is %ldx%ld  memory required is %.3fMB\n",
      (long int) width, (long int) height, (height * rowbytes + height *
      sizeof (unsigned char *) + width * height) / (1024.0 * 1024.0));

  image1 = (unsigned char *) malloc ((size_t) (height * rowbytes));
  row_pointers = (unsigned char **) malloc ((size_t) height *
    sizeof (unsigned char *));
  if (!image1 || !row_pointers)
    {
      fprintf (stderr, "pngrecolor error: unable to allocate memory for "
        "image\n");
      png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp) NULL);
      (void) fclose (fp);
      exit (4);
    }

  for (i = 0; i < (int) height; i++)
    row_pointers[i] = &image1[rowbytes * i];

  png_read_image (png_ptr, row_pointers);
  png_read_end (png_ptr, info_ptr);
  png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp) NULL);
  (void) fclose (fp);
}

static void
write_new_png (char *fn, int no_interlace, int verbose)
{
  png_structp png_ptr;
  png_infop info_ptr;
  png_color palette[256];
  unsigned char trans[256];
  FILE *fp;
  int num_trans;
  int i;
  png_color_16 newtrns;
  png_color_16 newbackground;

  memset (&newtrns, 0, sizeof (png_color_16));
  memset (&newbackground, 0, sizeof (png_color_16));

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
    {
      fprintf (stderr, "pngrecolor error: out of memory\n");
      exit (4);
    }

  fp = fopen (fn, "wb");
  if (!fp)
    {
      fprintf (stderr, "pngrecolor error: can't write to file %s\n", fn);
      png_destroy_write_struct (&png_ptr, NULL);
      exit (1);
    }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
    {
      fprintf (stderr, "pngrecolor error: can't create info struct\n");
      png_destroy_write_struct (&png_ptr, NULL);
      (void) fclose (fp);
      exit (4);
    }

  if (setjmp (png_jmpbuf (png_ptr)))
    {
      fprintf (stderr, "pngrecolor error: can't write file\n");
      png_destroy_write_struct (&png_ptr, NULL);
      (void) fclose (fp);
      exit (4);
    }

  png_init_io (png_ptr, fp);
  if (no_interlace)
    interlace_type = 0;

  if (valid_gray)
    {
      png_set_IHDR (png_ptr, info_ptr, width, height, new_bit_depth,
        PNG_COLOR_TYPE_GRAY, interlace_type, PNG_COMPRESSION_TYPE_BASE,
        PNG_FILTER_TYPE_BASE);
      if (gray_trns >= 0)
        {
          newtrns.gray = (png_uint_16) gray_trns;
          png_set_tRNS (png_ptr, info_ptr, NULL, 1, &newtrns);
        }
    }
  else
    {
      png_set_IHDR (png_ptr, info_ptr, width, height, new_bit_depth,
        PNG_COLOR_TYPE_PALETTE, interlace_type, PNG_COMPRESSION_TYPE_BASE,
        PNG_FILTER_TYPE_BASE);
      num_trans = 0;  /* number of transparent palette entries */
      for (i = 0; i < pal_used; i++)
        {
          palette[i].red = tmppal[i].red;
          palette[i].green = tmppal[i].green;
          palette[i].blue = tmppal[i].blue;
          trans[i] = tmppal[i].alpha;
          if (trans[i] < (unsigned char) 255)
            num_trans = i + 1;
        }
      png_set_PLTE (png_ptr, info_ptr, palette, pal_used);
      if (num_trans > 0)
        png_set_tRNS (png_ptr, info_ptr, trans, num_trans, 0);
    }

  if (chnks.has_gama)
    png_set_gAMA (png_ptr, info_ptr, image_gamma);
  if (chnks.has_srgb)
    png_set_sRGB (png_ptr, info_ptr, srgb_intent);
  if (chnks.has_bkgd)
    {
      if (valid_gray)
        newbackground.gray = (png_uint_16) bkgd_pal;
      else
        newbackground.index = bkgd_pal;
      png_set_bKGD (png_ptr, info_ptr, &newbackground);
    }

  png_write_info (png_ptr, info_ptr);
  png_set_packing (png_ptr);

  /* re-use row_pointers array */
  for (i = 0; i < (int) height; i++)
    row_pointers[i] = &image2[i * width];

  png_write_image (png_ptr, row_pointers);
  png_write_end (png_ptr, info_ptr);
  png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
  (void) fclose (fp);

  free (row_pointers);
  free (image2);
}


/* set ignorealpha to 1 if you don't care if the alpha value matches
 * (for the background) */
static int
add_to_palette (tmppal_t * pal, tmppal_t * color, int *pal_used,
  int ignorealpha, int verbose)
{
  int used;

  used = (*pal_used);
  if (used >= 256)
    {
      if (verbose)
        fprintf (stderr, "pngrecolor warning: more than 256 colors - no "
          "conversion\n");
      exit (3);
    }
  pal[used].red = color->red;
  pal[used].green = color->green;
  pal[used].blue = color->blue;
  pal[used].alpha = ignorealpha ? (unsigned char) 255 : color->alpha;
  pal[used].count = 1;
  (*pal_used)++;
  return 1;
}


/* Sort the palette to put transparent colors first, then by frequency,
 * then by dark to light on R+G+B, then G, then B, then R.
 * Sorting by frequency will often help the png filters make the image more
 * compressible.  It also makes it easier for people to see which colors aren't
 * used much and allow them to manually reduce the color palette. */
static int
palsortfunc (const void *p1, const void *p2)
{
  tmppal_t *e1, *e2;
  int s1, s2;

  e1 = (tmppal_t *) p1;
  e2 = (tmppal_t *) p2;

  if ((e1->alpha == (unsigned char) 255) && (e2->alpha < (unsigned char) 255))
    return (1);
  if ((e1->alpha < (unsigned char) 255) && (e2->alpha == (unsigned char) 255))
    return (-1);

  if (e1->count < e2->count)
    return (1);
  if (e1->count > e2->count)
    return (-1);

  s1 = (int) e1->red + (int) e1->green + (int) e1->blue;
  s2 = (int) e2->red + (int) e2->green + (int) e2->blue;
  if (s1 > s2)
    return (1);
  if (s1 < s2)
    return (-1);

  if (e1->green > e2->green)
    return (1);
  if (e1->green < e2->green)
    return (-1);
  if (e1->blue > e2->blue)
    return (1);
  if (e1->blue < e2->blue)
    return (-1);
  if (e1->red > e2->red)
    return (1);
  if (e1->red < e2->red)
    return (-1);

  return (0);
}


static int
find_pal_entry (unsigned char r, unsigned char g, unsigned char b,
  unsigned char a, int ignorealpha, int reset)
{
  int i;
  static int prev_entry;
  static int prev_r = (-1), prev_g = (-1), prev_b = (-1), prev_a = (-1);

  if (reset)
    {
      prev_r = prev_g = prev_b = prev_a = (-1);
      return -1;
    }

  if (prev_r == (int) r && prev_g == (int) g && prev_b == (int) b &&
      prev_a == (int) a)
    return prev_entry;

  for (i = 0; i < pal_used; i++)
    if (tmppal[i].red == r && tmppal[i].green == g && tmppal[i].blue == b
        && (tmppal[i].alpha == a || ignorealpha))
      {
        prev_r = (int) r;
        prev_g = (int) g;
        prev_b = (int) b;
        prev_a = (int) a;
        prev_entry = i;
        return i;
      }
  return (-1);
}


static void
make_new_png (int verbose)
{
  tmppal_t thispix;
  unsigned char *p;
  int i, palent, x, y;
  int gray_trns_palentry;                             /* temp palette entry */
  unsigned char gray_trns_shade = (unsigned char) 0;  /* 0 through 255 */

  pal_used = 0;

  for (y = 0; y < (int) height; y++)
    for (x = 0; x < (int) width; x++)
      {
        p = &row_pointers[y][x * channels];
        thispix.red = p[0];
        thispix.green = p[1];
        thispix.blue = p[2];
        if (channels == 4)
          thispix.alpha = p[3];
        else
          thispix.alpha = (unsigned char) 255;
        palent = find_pal_entry (p[0], p[1], p[2], thispix.alpha, 0, 0);
        if (palent < 0)
          {
            if (!add_to_palette (tmppal, &thispix, &pal_used, 0, verbose))
              {
                fprintf (stderr, "pngrecolor internal error: overloaded "
                  "palette\n");
                exit (4);
               }
          }
        else
          tmppal[palent].count++;
      }

  if (chnks.has_bkgd)
    {
      bkgd.alpha = (unsigned char) 255;  /* actually not needed */
      palent =
        find_pal_entry (bkgd.red, bkgd.green, bkgd.blue, bkgd.alpha, 1, 0);
      if (palent < 0)
        add_to_palette (tmppal, &bkgd, &pal_used, 1, verbose);
      else
        tmppal[palent].count++;
    }

  /* determine bit depth and whether we should do grayscale or palette */
  new_bit_depth = 8;
  if (pal_used <= 16)
    new_bit_depth = 4;
  if (pal_used <= 4)
    new_bit_depth = 2;
  if (pal_used <= 2)
    new_bit_depth = 1;

  /* figure out if this is a valid grayscale image */
  gray_trns_palentry = -1;
  valid_gray = 1;
  for (i = 0; i < pal_used; i++)
    {
      if (tmppal[i].red != tmppal[i].green || tmppal[i].red != tmppal[i].blue)
        {
          valid_gray = 0;
          break;
        }
      if (tmppal[i].alpha != (unsigned char) 255 &&
          tmppal[i].alpha != (unsigned char) 0)
        {
          valid_gray = 0;
          break;
        }
      if (tmppal[i].alpha == (unsigned char) 0)
        {
          if (gray_trns_palentry != -1)
            {
              valid_gray = 0;
              break;
            }  /* multiple transparent colors */
          gray_trns_palentry = i;  /* binary transparency ok (so far) */
          gray_trns_shade = tmppal[i].red;
        }
      switch (tmppal[i].red)
        {
        case 0:
        case 255:
          continue;
        case 85:
        case 170:
          if (new_bit_depth >= 2)
            continue;
          break;
        case 17:
        case 34:
        case 51:
        case 68:
        case 102:
        case 119:
        case 136:
        case 153:
        case 187:
        case 204:
        case 221:
        case 238:
          if (new_bit_depth >= 4)
            continue;
          break;
        default:
          if (new_bit_depth >= 8)
            continue;
        }
      valid_gray = 0;
      break;
    }

  /* One thing the above doesn't check for is a nontransparent
   * grayscale that's the same as the transparent grayscale color.
   * In this case we have to use palette color. */
  if (valid_gray && gray_trns_palentry != -1)
    if (find_pal_entry (gray_trns_shade, gray_trns_shade, gray_trns_shade,
        (unsigned char) 255, 0, 0) != -1)
      valid_gray = 0;

  /* put the palette in a good order */
  if (valid_gray)
    {
      /* If grayscale, create a "fake" palette consisting of all
       * available gray shades. */
      switch (new_bit_depth)
        {
        case 8:
          pal_used = 256;
          for (i = 0; i < pal_used; i++)
            {
              tmppal[i].red = tmppal[i].green = tmppal[i].blue =
                (unsigned char) i;
              tmppal[i].alpha = (unsigned char) 255;
            }
          break;
        case 4:
          pal_used = 16;
          for (i = 0; i < pal_used; i++)
            {
              tmppal[i].red = tmppal[i].green = tmppal[i].blue =
                (unsigned char) (i * 17);
              tmppal[i].alpha = (unsigned char) 255;
            }
          break;
        case 2:
          pal_used = 4;
          for (i = 0; i < pal_used; i++)
            {
              tmppal[i].red = tmppal[i].green = tmppal[i].blue =
                (unsigned char) (i * 85);
              tmppal[i].alpha = (unsigned char) 255;
            }
          break;
        case 1:
          pal_used = 2;
          for (i = 0; i < pal_used; i++)
            {
              tmppal[i].red = tmppal[i].green = tmppal[i].blue =
                (unsigned char) (i * 255);
              tmppal[i].alpha = (unsigned char) 255;
            }
          break;
        }

      gray_trns = -1;
      /* handle grayscale binary transparency */
      if (gray_trns_palentry != -1)
        {
          gray_trns = find_pal_entry (gray_trns_shade, gray_trns_shade,
            gray_trns_shade, (unsigned char) 255, 0, 0);
          if (gray_trns == -1)
            {
              fprintf (stderr, "pngrecolor internal error: can't find "
                "transparent grayscale color\n");
              exit (1);
            }
          tmppal[gray_trns].alpha = (unsigned char) 0;
        }
    }
  else
    qsort ((void *) tmppal, (size_t) pal_used, sizeof (tmppal_t), palsortfunc);

  if (verbose)
    {
      if (valid_gray)
        fprintf (stdout, "saving as grayscale:          %2d bpp\n",
          new_bit_depth);
      else
        fprintf (stdout, "new palette size:        %3d, %2d bpp\n", pal_used,
          new_bit_depth);
    }

  /* reset the find_pal_entry() cache */
  find_pal_entry ((unsigned char) 0, (unsigned char) 0, (unsigned char) 0,
    (unsigned char) 0, 0, 1);

  /* now create the new image */
  image2 = (unsigned char *) malloc ((size_t) (width * height));
  if (!image2)
    {
      fprintf (stderr, "pngrecolor error: out of memory\n");
      exit (4);
    }

  for (y = 0; y < (int) height; y++)
    for (x = 0; x < (int) width; x++)
      {
        p = &row_pointers[y][x * channels];
        palent = find_pal_entry (p[0], p[1], p[2],
          (unsigned char) ((channels == 4) ? p[3] : (unsigned char) 255),
          0, 0);
        if (palent < 0)
          {
            fprintf (stderr, "pngrecolor internal error: can't locate "
              "palette entry\n");
            exit (4);
          }
        image2[y * width + x] = (unsigned char) palent;
      }

  if (chnks.has_bkgd)
    {
      palent = find_pal_entry (bkgd.red, bkgd.green, bkgd.blue,
        (unsigned char) 255, 1, 0);
      if (palent < 0)
        {
          fprintf (stderr, "pngrecolor internal error: can't locale palette "
            "entry for bkgd\n");
          exit (4);
        }
      bkgd_pal = (unsigned char) palent;
    }
  free (image1);
}


static void
help (FILE *where)
{
  fprintf (where,
    "pngrecolor " PACKAGE_VERSION "\n"
    "usage: pngrecolor [-h(elp)] [-n(o_interlace)] [-q(uiet)] infile outfile\n");
}


int
main (int argc, char **argv)
{
  /* Define variables. */
  int c, no_interlace, verbose;

  /* Parse command-line options. */
  verbose = 1;
  no_interlace = 0;
  while ((c = getopt (argc, argv, "hnq")) != OPTEND)
    switch (c)
      {
      case 'h':
        help (stdout);
        return (0);
      case 'n':
        no_interlace = 0;
        break;
      case 'q':
        verbose = 0;
        break;
      case '?':
        help (stderr);
        return (1);
      }

  /* Ensure that two files were given as input. */
  if ((argc - optind) != 2)
    {
      fprintf (stderr, "pngrecolor error: missing input or output file\n");
      help (stderr);
      return (1);
    }

  /* Palette-optimize the image. */
  read_png (argv[optind], verbose);
  make_new_png (verbose);
  write_new_png (argv[optind + 1], no_interlace, verbose);

  /* Indicate successful finish. */
  return (0);
}
